package org.expeditee.items.widgets;

/*
 * JavaFX is not on the default java classpath until Java 8 (but is still included with Java 7), so your IDE will probably complain that the imports below can't be resolved. 
 * In Eclipse hitting'Proceed' when told 'Errors exist in project' should allow you to run Expeditee without any issues (although the JFX Browser widget will not display),
 * or you can just exclude JfxBrowser, WebParser and JfxbrowserActions from the build path.
 *
 * If you are using Ant to build/run, 'ant build' will try to build with JavaFX jar added to the classpath. 
 * If this fails, 'ant build-nojfx' will build with the JfxBrowser, WebParser and JfxbrowserActions excluded from the build path.
 */
import java.awt.Point;
import java.lang.reflect.Field;

import javafx.application.Platform;
import javafx.beans.value.ChangeListener;
import javafx.beans.value.ObservableValue;
import javafx.concurrent.Worker.State;
import javafx.embed.swing.JFXPanel;
import javafx.event.ActionEvent;
import javafx.event.Event;
import javafx.event.EventDispatchChain;
import javafx.event.EventDispatcher;
import javafx.event.EventHandler;
import javafx.geometry.Point2D;
import javafx.geometry.Rectangle2D;
import javafx.scene.Node;
import javafx.scene.Scene;
import javafx.scene.control.Button;
import javafx.scene.control.TextField;
import javafx.scene.input.KeyEvent;
import javafx.scene.input.MouseButton;
import javafx.scene.input.MouseEvent;
import javafx.scene.layout.HBox;
import javafx.scene.layout.Priority;
import javafx.scene.layout.VBox;
import javafx.scene.web.WebEngine;
import javafx.scene.web.WebView;
import netscape.javascript.JSObject;

import org.expeditee.gui.DisplayIO;
import org.expeditee.gui.FrameMouseActions;
import org.expeditee.gui.FreeItems;
import org.expeditee.gui.MessageBay;
import org.expeditee.io.WebParser;
import org.expeditee.items.Item;
import org.expeditee.items.Picture;
import org.expeditee.items.Text;
import org.expeditee.settings.network.NetworkSettings;

import com.sun.javafx.scene.control.skin.TextFieldSkin;
import com.sun.javafx.scene.text.HitInfo;

/**
 * Web browser using a JavaFX WebView.
 * 
 * @author ngw8
 * @author jts21
 */
public class JfxBrowser extends DataFrameWidget {
	
	private static final String BACK = "back";
	private static final String FORWARD = "forward";
	private static final String REFRESH = "refresh";
	private static final String CONVERT = "convert";
	
	private JFXPanel _panel;
	private WebView _webView;
	private WebEngine _webEngine;
	private TextField _urlField;
	
	private MouseButton _buttonDownId = MouseButton.NONE;
	private MouseEvent _backupEvent = null;
	private static Field MouseEvent_x, MouseEvent_y;

	static {
		Platform.setImplicitExit(false);
		
		try {
	        MouseEvent_x = MouseEvent.class.getDeclaredField("x");
    		MouseEvent_x.setAccessible(true);
    		MouseEvent_y = MouseEvent.class.getDeclaredField("y");
    		MouseEvent_y.setAccessible(true);
		} catch (Exception e) {
	        e.printStackTrace();
        }
	}
	
	private Point getCoordFromCaret(TextField text) {
		TextFieldSkin skin = (TextFieldSkin) text.getSkin();
		
		Point2D onScene = text.localToScene(0, 0);
		
		double x = onScene.getX() + JfxBrowser.this.getX();// - org.expeditee.gui.Browser._theBrowser.getOrigin().x;
		double y = onScene.getY() + JfxBrowser.this.getY();// - org.expeditee.gui.Browser._theBrowser.getOrigin().y;
		
		Rectangle2D cp = skin.getCharacterBounds(text.getCaretPosition());
		
		return new Point((int) (cp.getMinX() + x), (int) (cp.getMinY() + y));
	}
	
	private int getCaretFromCoord(TextField text, MouseEvent e) {
		TextFieldSkin skin = (TextFieldSkin) text.getSkin();
		HitInfo hit = skin.getIndex(e);
		return hit.getInsertionIndex();
	}
	
	/**
	 * @param src The MouseEvent to clone
	 * @param node The node the position will be relative to
	 * @param x The position in Expeditee space
	 * @param y The position in Expeditee space
	 * @return A fake MouseEvent for a specific position relative to a Node
	 */
	private MouseEvent getMouseEventForPosition(MouseEvent src, Node node, int x, int y) {
		MouseEvent dst = (MouseEvent) ((Event) src).copyFor(null, null);
		try {
	        MouseEvent_x.set(dst, x - node.localToScene(0, 0).getX());
	        MouseEvent_y.set(dst, y - node.localToScene(0, 0).getY());
        } catch (Exception e) {
	        e.printStackTrace();
        }
		return dst;
	}

	public JfxBrowser(Text source, final String[] args) {
		// Initial page is either the page stored in the arguments (if there is one stored) or the homepage
		super(source, new JFXPanel(), -1, 500, -1, -1, 300, -1);

		_panel = (JFXPanel) _swingComponent;
		
		Platform.runLater(new Runnable() {
			@Override
			public void run() {
				initFx((args != null && args.length > 0) ? args[0] : NetworkSettings.HomePage.get());
			}
		});
	}
	
	/**
	 * Sets up the browser frame. JFX requires this to be run on a new thread.
	 * 
	 * @param url
	 *            The URL to be loaded when the browser is created
	 */
	private void initFx(String url) {
		try {
			VBox vertical = new VBox();
			HBox horizontal = new HBox();
			
			Button backButton = new Button("Back");
			backButton.setMinWidth(Button.USE_PREF_SIZE);
			backButton.setFocusTraversable(false);
			Button forwardButton = new Button("Forward");
			forwardButton.setMinWidth(Button.USE_PREF_SIZE);
			forwardButton.setFocusTraversable(false);
			this._urlField = new TextField(url);
			this._urlField.setMinWidth(0);
			this._urlField.setMaxWidth(Double.MAX_VALUE);
			this._urlField.setFocusTraversable(false);
			Button convertButton = new Button("Convert");
			convertButton.setMinWidth(Button.USE_PREF_SIZE);
			convertButton.setFocusTraversable(false);
			Button convertButtonNew = new Button("Convert (New Method)");
			convertButtonNew.setMinWidth(Button.USE_PREF_SIZE);
			convertButtonNew.setFocusTraversable(false);
			
			horizontal.getChildren().addAll(backButton, forwardButton, this._urlField, convertButton, convertButtonNew);
			
			HBox.setHgrow(backButton, Priority.NEVER);
			HBox.setHgrow(forwardButton, Priority.NEVER);
			HBox.setHgrow(convertButton, Priority.NEVER);
			HBox.setHgrow(convertButtonNew, Priority.NEVER);
			HBox.setHgrow(this._urlField, Priority.ALWAYS);
			
			this._webView = new WebView();
			this._webView.setMaxWidth(Double.MAX_VALUE);
			this._webView.setMaxHeight(Double.MAX_VALUE);
			HBox.setHgrow(this._webView, Priority.ALWAYS);
			VBox.setVgrow(this._webView, Priority.ALWAYS);
			this._webEngine = this._webView.getEngine();
			
			vertical.getChildren().addAll(horizontal, this._webView);
			
			Scene scene = new Scene(vertical);

			this._panel.setScene(scene);
			
			// Disable right click menu
			this._webView.setContextMenuEnabled(false);
			final EventDispatcher initial = this._urlField.getEventDispatcher();
			this._urlField.setEventDispatcher(new EventDispatcher() {
				@Override
				public Event dispatchEvent(Event e, EventDispatchChain tail) {
					if (e instanceof MouseEvent) {
						MouseEvent m = (MouseEvent) e;
                        if (m.getButton() == MouseButton.SECONDARY && m.getEventType() == MouseEvent.MOUSE_RELEASED) {
                            e.consume();
                            JfxBrowser.this._urlField.getOnMouseReleased().handle(m);
                        }
                    }
                    return initial.dispatchEvent(e, tail);
				}
			});
			
			backButton.setOnAction(new EventHandler<ActionEvent>() {
				@Override
                public void handle(ActionEvent e) {
                    navigateBack();
                }
			});

			forwardButton.setOnAction(new EventHandler<ActionEvent>() {
				@Override
                public void handle(ActionEvent e) {
                    navigateForward();
                }
			});
			
			
			convertButton.setOnAction(new EventHandler<ActionEvent>() {
				@Override
                public void handle(ActionEvent e) {
                    getFrame();
                }
			});

			convertButtonNew.setOnAction(new EventHandler<ActionEvent>() {
				@Override
                public void handle(ActionEvent e) {
                    getFrameNew();
                }
			});

			this._urlField.setOnAction(new EventHandler<ActionEvent>() {
				@Override
                public void handle(ActionEvent e) {
                    navigate(JfxBrowser.this._urlField.getText());
                }
			});

			this._urlField.setOnKeyTyped(new EventHandler<KeyEvent>() {
				@Override
				public void handle(KeyEvent e) {
					// Hiding the cursor when typing, to be more Expeditee-like
					DisplayIO.setCursor(org.expeditee.items.Item.HIDDEN_CURSOR);
				}
			});

			this._urlField.setOnMouseMoved(new EventHandler<MouseEvent>() {
				@Override
				public void handle(MouseEvent e) {
					JfxBrowser.this._backupEvent = e;
					// make sure we have focus if the mouse is moving over us
					if(!JfxBrowser.this._urlField.isFocused()) {
						JfxBrowser.this._urlField.requestFocus();
					}
					// Checking if the user has been typing - if so, move the cursor to the caret position
					if (DisplayIO.getCursor() == Item.HIDDEN_CURSOR) {
						DisplayIO.setCursor(org.expeditee.items.Item.TEXT_CURSOR);
						DisplayIO.setCursorPosition(getCoordFromCaret(JfxBrowser.this._urlField));
					} else {
						// Otherwise, move the caret to the cursor location
						// int x = FrameMouseActions.getX() - JfxBrowser.this.getX(), y = FrameMouseActions.getY() - JfxBrowser.this.getY();
						JfxBrowser.this._urlField.positionCaret(getCaretFromCoord(JfxBrowser.this._urlField, e));
					}
				}
			});
			
			this._urlField.setOnMouseEntered(new EventHandler<MouseEvent>() {
				@Override
                public void handle(MouseEvent arg0) {
					JfxBrowser.this._urlField.requestFocus();
                }
			});
			
			this._urlField.setOnMouseExited(new EventHandler<MouseEvent>() {
				@Override
                public void handle(MouseEvent arg0) {
					JfxBrowser.this._webView.requestFocus();
                }
			});
			
			this._urlField.setOnMouseDragged(new EventHandler<MouseEvent>() {
				@Override
				public void handle(MouseEvent e) {
					TextFieldSkin skin = (TextFieldSkin) JfxBrowser.this._urlField.getSkin();
					
					if (!JfxBrowser.this._urlField.isDisabled()) {
                        if (JfxBrowser.this._buttonDownId == MouseButton.MIDDLE || JfxBrowser.this._buttonDownId == MouseButton.SECONDARY) {
                            if (!(e.isControlDown() || e.isAltDown() || e.isShiftDown() || e.isMetaDown())) {
                                skin.positionCaret(skin.getIndex(e), true);
                            }
                        }
                    }
				}
			});
			
			this._urlField.focusedProperty().addListener(new ChangeListener<Boolean>() {
				@Override
                public void changed(ObservableValue<? extends Boolean> property, Boolean oldValue, Boolean newValue) {
					if(newValue.booleanValue()) {
						DisplayIO.setCursor(org.expeditee.items.Item.TEXT_CURSOR);
					} else {
						// Restoring the standard cursor, since it is changed to a text cursor when focus is gained
						DisplayIO.setCursor(org.expeditee.items.Item.DEFAULT_CURSOR);
					}
                }
			});

			this._urlField.setOnMouseReleased(new EventHandler<MouseEvent>() {
				@Override
                public void handle(MouseEvent e) {
                    JfxBrowser.this._buttonDownId = MouseButton.NONE;

					Text item;

					// If nothing is selected, then select all the text so that it will be copied/moved
					if (JfxBrowser.this._urlField.getSelectedText() == null || JfxBrowser.this._urlField.getSelectedText().length() == 0) {
						JfxBrowser.this._urlField.selectAll();
					}

					if (e.getButton() == MouseButton.SECONDARY) {
						// Right mouse button released, so copy the selection (i.e. don't remove the original)
						item = DisplayIO.getCurrentFrame().createNewText(JfxBrowser.this._urlField.getSelectedText());
						FrameMouseActions.pickup(item);
					} else if (e.getButton() == MouseButton.MIDDLE) {
						// Middle mouse button released, so copy the selection then remove it from the URL field
						item = DisplayIO.getCurrentFrame().createNewText(JfxBrowser.this._urlField.getSelectedText());
						JfxBrowser.this._urlField.setText(
								JfxBrowser.this._urlField.getText().substring(0, JfxBrowser.this._urlField.getSelection().getStart())
								+ JfxBrowser.this._urlField.getText().substring(JfxBrowser.this._urlField.getSelection().getEnd(),
										JfxBrowser.this._urlField.getText().length()));

						FrameMouseActions.pickup(item);
					}
                }
			});
			
			this._urlField.setOnMousePressed(new EventHandler<MouseEvent>() {
				@Override
                public void handle(MouseEvent e) {
					JfxBrowser.this._buttonDownId = e.getButton();
				}
			});

			this._webEngine.getLoadWorker().stateProperty().addListener(new ChangeListener<State>() {

				@Override
				public void changed(ObservableValue<? extends State> ov, State oldState, State newState) {

					switch (newState) {
					case READY: // READY
						// MessageBay.displayMessage("WebEngine ready");
						break;
					case SCHEDULED: // SCHEDULED
						// MessageBay.displayMessage("Scheduled page load");
						break;
					case RUNNING: // RUNNING
						System.out.println("Loading page!");

						// Updating the URL bar to display the URL of the page being loaded
						JfxBrowser.this._urlField.setText(JfxBrowser.this._webEngine.getLocation());

						break;
					case SUCCEEDED: // SUCCEEDED
						MessageBay.displayMessage("Finished loading page");
						break;
					case CANCELLED: // CANCELLED
						MessageBay.displayMessage("Cancelled loading page");
						break;
					case FAILED: // FAILED
						MessageBay.displayMessage("Failed to load page");
						break;
					}
				}
			});
			
			// Captures mouse click events on webview to enable expeditee like behavior for JavaFX browser.
			this._webView.setOnMouseClicked(new EventHandler<javafx.scene.input.MouseEvent>() {
				@Override
				public void handle(javafx.scene.input.MouseEvent e) {
					if(e.getButton() == MouseButton.SECONDARY) {
						// Gets text currently selected in webview
						String selection = (String) JfxBrowser.this._webEngine.executeScript("window.getSelection().toString()");

						// If no text is selected, see if an image is under the cursor
						if (selection.length() == 0) {
							JSObject window = (JSObject) JfxBrowser.this._webEngine.executeScript("window");
							Object image = JfxBrowser.this._webEngine.executeScript(""
									+ "function isImage(o) {"
									+ "  if(o.tagName == \"IMG\" && o.src != null) {"
									+ "    return true;"
									+ "  }"
									+ "  if(window.getComputedStyle(o).getPropertyValue(\"background-image\").indexOf(\"url\") == 0) {"
									+ "    return true;"
									+ "  }"
									+ "  return false;"
									+ "};"
									+ "var clicked = document.elementFromPoint(" + e.getX() + "," + e.getY() + ");"
									+ "if(isImage(clicked)) {"
									+ "  clicked;"
									+ "} else {"
									// TODO: look in child elements and other elements behind the current one
									// For now just return nothing
									+ "  null;"
									+ "}"
									+ "");
							if(image instanceof org.w3c.dom.Node) {
								System.out.println(image);
								try {
									JSObject style = (JSObject) window.call("getComputedStyle", image);

									JSObject bounds = (JSObject) ((JSObject) image).call("getBoundingClientRect", new Object[] {});
									float width = Float.valueOf(bounds.getMember("width").toString());
									float height = Float.valueOf(bounds.getMember("height").toString());
									
									Picture pic;
									
									org.w3c.dom.Node node = ((org.w3c.dom.Node) image);
									
									if (((String) style.call("getPropertyValue", new Object[] { "background-image" })).startsWith("url(")) {
										pic = WebParser.getBackgroundImageFromNode(node, style, DisplayIO.getCurrentFrame(), null,
											(float) FrameMouseActions.getX(), (float) FrameMouseActions.getY(), width, height);
									} else {
										String imgSrc;
										if(node.getNodeName().toLowerCase().equals("img") && 
												(imgSrc = ((JSObject) node).getMember("src").toString()) != null) {
											pic = WebParser.getImageFromUrl(imgSrc, null, DisplayIO.getCurrentFrame(),
													(float) FrameMouseActions.getX(), (float) FrameMouseActions.getY(), (int) width, null, null, null, null, null, 0, 0);
										} else {
											return;
										}
									}
									pic.setXY(FrameMouseActions.getX(), FrameMouseActions.getY());
                            		FrameMouseActions.pickup(pic);
                        		} catch (Exception e1) {
	                                // TODO Auto-generated catch block
	                                e1.printStackTrace();
                                }
							}
						} else {
							// Copy text and attach to cursor
							Text t = new Text(selection);
							t.setParent(DisplayIO.getCurrentFrame());
							t.setXY(FrameMouseActions.getX(), FrameMouseActions.getY());
							FrameMouseActions.pickup(t);
						}
					}
				}
			});

			this._webEngine.load(url);
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	public void navigate(String url) {
		final String actualURL;
		String urlLower = url.toLowerCase();
		// check if protocol is missing
		if (!(urlLower.startsWith("http://") || url.startsWith("https://") || urlLower.startsWith("ftp://") || urlLower.startsWith("file://"))) {
			// check if it's a search
			int firstSpace = url.indexOf(" ");
			int firstDot = url.indexOf(".");
			int firstSlash = url.indexOf('/');
			int firstQuestion = url.indexOf('?');
			int firstSQ;
			if(firstSlash == -1) {
				firstSQ = firstQuestion;
			} else if(firstQuestion == -1) {
				firstSQ = firstSlash;
			} else {
				firstSQ = -1;
			}
			if(firstDot <= 0 ||                                        // no '.' or starts with '.' -> search
			        (firstSpace != -1 && firstSpace < firstDot + 1) || // ' ' before '.'            -> search
			        (firstSpace != -1 && firstSpace < firstSQ)) {      // no '/' or '?'             -> search
				// make it a search
				actualURL = NetworkSettings.SearchEngine.get() + url;
			} else {
				// add the missing protocol
				actualURL = "http://" + url;
			}
		} else {
			actualURL = url;
		}
		System.out.println(actualURL);
		try {
			Platform.runLater(new Runnable() {
				@Override
				public void run() {
					try {
						JfxBrowser.this._webEngine.load(actualURL);
					} catch (Exception e) {
						e.printStackTrace();
					}
				}
			});
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	/**
	 * Navigates JfxBrowser back through history. If end of history reached the user is notified via the MessageBay.
	 * Max size of history is 100 by default.
	 */
	public void navigateBack() {
		try {
		Platform.runLater(new Runnable() {
			@Override
			public void run() {
				try {
					JfxBrowser.this._webEngine.getHistory().go(-1);
					FreeItems.getInstance().clear();
				} catch (IndexOutOfBoundsException e) {
					MessageBay.displayMessage("Start of History");
					}
			}
		});
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
	
	/**
	 * Navigates JfxBrowser forward through history. If end of history reached the user is notified via the MessageBay.
	 * Max size of history is 100 by default.
	 */
	public void navigateForward() {
		try {
		Platform.runLater(new Runnable() {
			@Override
			public void run() {
				try {
					JfxBrowser.this._webEngine.getHistory().go(1);
					FreeItems.getInstance().clear();
				} catch (IndexOutOfBoundsException e) {
					MessageBay.displayMessage("End of History");
					}
			}
		});
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
	
	/**
	 * Refreshes webview by reloading the page.
	 */
	public void refresh() {
		try {
		Platform.runLater(new Runnable() {
			@Override
			public void run() {
				try {
					JfxBrowser.this._webEngine.reload();
					FreeItems.getInstance().clear();
					MessageBay.displayMessage("Page Reloading");
				} catch (Exception e) {
					e.printStackTrace();
				}
			}
		});
		} catch (Exception e) {
			e.printStackTrace();
		}

	}
	
	/**
	 * Traverses DOM an turns elements into expeditee items.
	 */
	public void getFrame() {
		try {
			WebParser.parsePage(this._webEngine, DisplayIO.getCurrentFrame());
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	public void getFrameNew() {
		// this.setSize(1000, 1000);
		// this._browser.setBounds(getX(), getY(), getWidth(), getHeight());
		// this.layout(this._browser);

		try {
			// hack to make sure we don't try parsing the page from within the JavaFX thread,
			// because doing so causes deadlock
			new Thread(new Runnable() {
				public void run() {
					WebParser.parsePageSimple(JfxBrowser.this._webEngine, JfxBrowser.this._webView, DisplayIO.getCurrentFrame());
				}
			}).start();
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	/**
	 * Used to drop text items onto JfxBrowser widget. Does nothing if a text item is not attached to cursor. <br>
	 * "back" -> navigates back a page in browser's session history <br>
	 * "forward" -> navigates forward a page in browser's session history <br>
	 * "refresh" -> reloads current page <br>
	 * "getFrame" -> attempts to parse page into an expeditee frame <br>
	 * url -> all other text is assumed to be a url which browser attempts to navigate to
	 * 
	 * @return Whether a JfxBrowser specific event is run.
	 * 
	 */
	@Override
	public boolean ItemsLeftClickDropped() {
		Text carried = null;
		if ((carried = FreeItems.getTextAttachedToCursor()) == null) {	// fails if no text is attached to cursor.
			return false;
		}
		
		if(carried.getText().toLowerCase().equals(BACK)) {
			navigateBack();
		} else if(carried.getText().toLowerCase().equals(FORWARD)) {
			navigateForward();
		} else if(carried.getText().toLowerCase().equals(REFRESH)) {
			refresh();
		} else if(carried.getText().toLowerCase().equals(CONVERT)) {
			getFrame();
		} else {
			String text = carried.getText().trim();
			this.navigate(text);
			FreeItems.getInstance().clear();
		}
		
		return true;
	}

	/**
	 * Used to enable expeditee like text-widget interaction for middle mouse clicks. Does nothing if a text item is not attached to cursor.
	 * @return false if a text-widget interaction did not occur, true if a text-widget interaction did occur.
	 */
	@Override
	public boolean ItemsMiddleClickDropped() {
		if(ItemsRightClickDropped()) {
			FreeItems.getInstance().clear();						// removed held text item - like normal expeditee middle click behaviour.
			return true;
		}
		return false;
	}
	
	/**
	 * Used to enable expeditee like text-widget interaction for right mouse clicks. Does nothing if a text item is not attached to cursor.
	 * @return false if a text-widget interaction did not occur, true if a text-widget interaction did occur.
	 */
	@Override
	public boolean ItemsRightClickDropped() {
		Text t = null;
		if((t = FreeItems.getTextAttachedToCursor()) == null) {	// fails if no text item is attached to the cursor.
			return false;
		}

		final int x = FrameMouseActions.getX() - this.getX(), y = FrameMouseActions.getY() - this.getY();
		if(!this._urlField.getBoundsInParent().contains(x, y)) {
			// fails if not clicking on urlField
			return false;
		}

		final String insert = t.getText();
		Platform.runLater(new Runnable() {
			@Override
			public void run() {
				// Inserts text in text item into urlField at the position of the mouse.
        		String s = JfxBrowser.this._urlField.getText();
        		int index = getCaretFromCoord(JfxBrowser.this._urlField, getMouseEventForPosition(JfxBrowser.this._backupEvent, JfxBrowser.this._urlField, x, y));
        		if(index < s.length()) {
        			s = s.substring(0, index) + insert + s.substring(index);
        		} else {
        			s = s + insert;
        		}
        		JfxBrowser.this._urlField.setText(s);
			}
		});

		return true;
	}
	
	@Override
	protected String[] getArgs() {
		String[] r = null;
		if (this._webView != null) {
			try {
				r = new String[] { this._webEngine.getLocation() };
			} catch (Exception e) {
				e.printStackTrace();
			}
		}
		return r;
	}
}
